import pandas as pd
from interface import Interface
from typing import Tuple, List, Dict, Optional
from datetime import date

from fast_tradier.models.trading.OptionOrder import OptionOrder # order for placing
from fast_tradier.models.trading.EquityOrder import EquityOrder
from fast_tradier.models.account.AccountOrder import AccountOrder # retrieved from account
from fast_tradier.models.market_data.Quote import Quote
from fast_tradier.models.account.Position import Position
from fast_tradier.models.trading.Interval import Interval
from fast_tradier.models.account.AccountBalance import AccountBalance
from fast_tradier.models.trading.OrderBase import OrderBase

class IBrokerAsyncClient(Interface):

    async def is_market_in_session_now_async(self) -> Tuple:
        raise NotImplementedError('not implemented in interface')
    
    async def is_market_open_today_async(self, market: str='NYSE') -> Tuple:
        raise NotImplementedError('not implemented in interface')

    async def get_quotes_async(self, symbols: List[str]) -> List[Quote]:
        raise NotImplementedError('not implemented in interface')

    def get_history_async(self, symbol: str, start_date: date, end_date: date, interval: Interval = Interval.Daily, open_hours_only: bool = True) -> Optional[pd.DataFrame]:
        """Get history data for the given ticker symbol. start_date and end_date must be in date type.
        Returns a dataframe, with columns: open, high, low, close, volume. Index is their date.
        """
        raise NotImplemented('not implemented in interface')
    
    async def get_order_status_async(self, order_id: int) -> Optional[str]:
        raise NotImplementedError('not implemented in interface')
    
    async def get_option_expirations_async(self, symbol: str) -> List[str]:
        raise NotImplementedError('not implemented in interface')
    
    async def place_order_async(self, order: OrderBase) -> Optional[int]:
        '''place order of any type: option or equity'''
        raise NotImplementedError('not implemented in interface')

    async def place_option_order_async(self, order: OptionOrder) -> Optional[int]:
        raise NotImplementedError('not implemented in interface')
    
    async def place_equity_order_async(self, order: EquityOrder) -> Optional[int]:
        raise NotImplementedError('not implemented in interface')

    async def cancel_order_async(self, order_id: int) -> bool:
        raise NotImplementedError('not implemented in interface')

    async def get_option_chain_async(self, symbol: str, expiration: str, greeks: bool = True) -> Optional[Dict]:
        raise NotImplementedError('not implemented in interface')

    async def modify_option_order_async(self, modified_order: OptionOrder) -> bool:
        raise NotImplementedError('not implemented in interface')
    
    async def get_positions_async(self) -> List[Position]:
        raise NotImplementedError('not implemented in interface')
    
    async def get_account_balance_async(self) -> Optional[AccountBalance]:
        raise NotImplementedError('not implemented in interface')

    async def get_account_orders_async(self) -> List[AccountOrder]:
        raise NotImplementedError('not implemented in interface')


from interface import Interface
from typing import Optional

class IRealTimeQuoteProvider(Interface):
    '''interface for getting real time quote/price for a symbol'''
    def get_price(self, symbol: str) -> Optional[float]:
        raise NotImplementedError('not implemented in interface')


from dataclasses import dataclass
from typing import Dict, Optional, Mapping, Any
from datetime import datetime
from pydantic import ConfigDict, Field
from fast_tradier.utils.TimeUtils import TimeUtils
from fast_tradier.models.ModelBase import NewModelBase
from fast_tradier.models.DataClassModelBase import DataClassModelBase

class Quote(NewModelBase):
    symbol: str
    description: Optional[str] = None
    exch: Optional[str] = None
    type_: str = Field(..., alias="type")
    last: Optional[float] = None
    change: Optional[float] = None
    volume: Optional[float] = None
    open_price: Optional[float] = Field(..., alias="open")
    high: Optional[float] = None
    low: Optional[float] = None
    close: Optional[float] = None
    bid: Optional[float] = None
    ask: Optional[float] = None
    underlying: Optional[str] = None
    strike: Optional[float] = None
    change_percentage: Optional[float] = None
    average_volume: Optional[int] = None
    last_volume: Optional[int] = None
    trade_date: Optional[int] = None
    prevclose: Optional[float] = None
    week_52_high: Optional[float] = None
    week_52_low: Optional[float] = None
    bidsize: Optional[int] = None
    bidexch: Optional[str] = None
    bid_date: Optional[int] = None
    asksize: Optional[int] = None
    askexch: Optional[str] = None
    ask_date: Optional[int] = None
    open_interest: Optional[int] = None
    contract_size: Optional[int] = None
    expiration_date: Optional[str] = None
    expiration_type: Optional[str] = None
    option_type: Optional[str] = None
    root_symbol: Optional[str] = None
    model_config = ConfigDict(use_enum_values=True, populate_by_name=True)

    @property
    def is_option(self) -> bool:
        return self.type.upper() == 'OPTION'

    @property
    def is_stock(self) -> bool:
        return self.type.upper() == 'STOCK'

    @property
    def trade_date_datetime(self) -> Optional[datetime]:
        return TimeUtils.convert_unix_ts(self.trade_date)

    @property
    def bid_date_datetime(self) -> Optional[datetime]:
        return TimeUtils.convert_unix_ts(self.bid_date)

    @property
    def ask_date_datetime(self) -> Optional[datetime]:
        return TimeUtils.convert_unix_ts(self.ask_date)


from fast_tradier.models.ModelBase import ModelBase
from typing import List, Dict, Optional
from fast_tradier.models.trading.Sides import EquityOrderSide
from fast_tradier.models.trading.PriceTypes import EquityPriceType
from fast_tradier.models.trading.Duration import Duration
from fast_tradier.models.trading.OrderBase import OrderBase
from pydantic import Field

class EquityOrder(OrderBase):
    clazz: str = Field(..., alias="class")
    symbol: str
    side: EquityOrderSide
    quantity: float
    duration: Duration
    price: Optional[float] = None
    price_type: EquityPriceType = Field(..., alias="type")
    stop: Optional[float] = None
    tag: Optional[str] = None

    @property
    def ticker(self) -> str:
        return self.symbol


from typing import Dict, List, Optional, Mapping, Any
from enum import Enum
from fast_tradier.models.trading.TOSTradierConverter import TOSTradierConverter
from fast_tradier.models.ModelBase import ModelBase
from fast_tradier.models.trading.Sides import OptionOrderSide
from fast_tradier.models.trading.PriceTypes import OptionPriceType
from fast_tradier.models.trading.Duration import Duration
from fast_tradier.models.trading.OrderBase import OrderBase
from fast_tradier.models.ModelBase import NewModelBase
from pydantic import BaseModel, ConfigDict, Field

class OptionLeg(NewModelBase):
    underlying_symbol: Optional[str] = None
    option_symbol: str
    quantity: int
    side: OptionOrderSide
    model_config = ConfigDict(use_enum_values=True, populate_by_name=True)

    def reverse_side(self) -> None:
        if self.side == OptionOrderSide.SellToOpen or self.side == OptionOrderSide.SellToOpen.value:
            self.side = OptionOrderSide.BuyToClose.value
        elif self.side == OptionOrderSide.BuyToOpen or self.side == OptionOrderSide.BuyToOpen.value:
            self.side = OptionOrderSide.SellToClose.value

class OptionOrderClass(str, Enum):
    SingleLeg = "option"
    MultiLegs = "multileg"

class OptionOrder(NewModelBase):
    symbol: str
    price: Optional[float] = None
    stop: Optional[float] = None
    price_type: OptionPriceType = Field(..., alias="type")
    duration: Duration
    option_legs: List[OptionLeg]
    side: Optional[OptionOrderSide] = None # required for single leg option order
    status: Optional[str] = "pending"
    id: Optional[int] = None
    tag: Optional[str] = None
    model_config = ConfigDict(use_enum_values=True, populate_by_name=True)

    @property
    def ticker(self) -> str:
        """Convenience prop to get underlying symbol."""
        return self.symbol
    
    @property
    def order_class(self) -> Optional[str]:
        if len(self.option_legs) == 1:
            return OptionOrderClass.SingleLeg.value
        elif len(self.option_legs) > 1:
            return OptionOrderClass.MultiLegs.value

        return None

    def clone_option_legs(self, reverse_side: bool = False) -> List[OptionLeg]:
        '''deep clone option_legs'''
        cloned_legs = []
        for opt_leg in self.option_legs:
            leg = OptionLeg(**(opt_leg.to_json()))
            if reverse_side:
                leg.reverse_side()

            cloned_legs.append(leg)
        return cloned_legs
    
    def to_json(self) -> Mapping[str, Any]:
        """
        Override to_json so that the format complies to Tradier API.
        Returns:
            Mapping[str, Any]: _description_
        """
        temp_result = self.model_dump(by_alias=True)
        result: Dict[str, Any] = {}
        for k, v in temp_result.items():
            result[k] = v
        if self.option_legs is not None and len(self.option_legs) > 0:
            result["class"] = self.order_class
            if len(self.option_legs) == 1:
                result["option_symbol"] = self.option_legs[0].option_symbol
                result["side"] = self.option_legs[0].side
                result["quantity"] = self.option_legs[0].quantity
            elif len(self.option_legs) > 1:
                for i in range(len(self.option_legs)):
                    opt_item = self.option_legs[i]
                    symbol_key = f"option_symbol[{i}]"
                    result[symbol_key] = opt_item.option_symbol
                    side_key = f"side[{i}]"
                    result[side_key] = f"{opt_item.side}"
                    quant_key = f"quantity[{i}]"
                    result[quant_key] = f"{opt_item.quantity}"
        else:
            raise Exception("option legs should not be empty")

        return result


from enum import Enum

class EquityOrderSide(str, Enum):
    Buy = "buy"
    BuyToCover = "buy_to_cover"
    Sell = "sell"
    SellShort = "sell_short"

class OptionOrderSide(str, Enum):
    BuyToOpen = "buy_to_open"
    BuyToClose = "buy_to_close"
    SellToOpen = "sell_to_open"
    SellToClose = "sell_to_close"


import re
from typing import Optional

tradier_option_symbol_pattern = re.compile(r'^([A-Z]+)([0-9]{2})([0-9]{4})(C|P)0?(\d+)0{3}')
class TOSTradierConverter(object):
    '''utility class for converting option symbols from Tradier to TOS or vice versa'''
    @staticmethod
    def tos_to_tradier(tos_option_symbol: str) -> Optional[str]:
        # e.g. SPXW230522C04210000
        tradier_option_format = '{ticker}{yy}{mmdd}{opt_type}{strike}000' # with padded 0s
        # e.g. SPXW_052523P1200
        tos_option_symbol_pattern = re.compile(r'^([A-Z]+)_([0-9]{4})([0-9]{2})(C|P)((\d)+)')
        # tradier_option_pattern = re.compile('^([A-Z]+)([2-9]{2})([0-9]{4})(C|P)0((\d)+)')

        d = re.search(tos_option_symbol_pattern, tos_option_symbol.upper())
        if d:
            ticker = d.group(1)
            mmdd = d.group(2)
            yy = d.group(3)
            option_type = d.group(4)
            strike = d.group(5)
            if len(strike) > 5:
                result = result[ : 5-len(strike)]
            elif len(strike) < 5:
                while len(strike) < 5:
                    strike = '0' + strike #prepend leading zeros to strike
                    result = tradier_option_format.format(ticker=ticker, yy=yy, mmdd=mmdd, opt_type=option_type, strike=strike)

            return result

        return tos_option_symbol

    @staticmethod
    def tradier_to_tos(tradier_option_symbol: str) -> str:
        d = re.search(tradier_option_symbol_pattern, tradier_option_symbol.upper())
        if d:
            ticker = d.group(1)
            yy = d.group(2)
            mmdd = d.group(3)
            opt_type = d.group(4)
            strike = d.group(5).lstrip('0') # remove leading zeroes
            tos_option_symbol = f'{ticker}_{mmdd}{yy}{opt_type}{strike}'
            return tos_option_symbol

        return tradier_option_symbol

    @staticmethod
    def get_strike(tradier_option_symbol: Optional[str]) -> Optional[float]:
        '''Parse out strike in float type from Tradier option symbol, could be None.'''
        if tradier_option_symbol is None:
            return None

        d = re.search(tradier_option_symbol_pattern, tradier_option_symbol.upper())
        if d:
            strike = d.group(5).lstrip('0') # remove leading zeroes
            try:
                return float(strike)
            except:
                return None


import json
from dataclasses import dataclass
from typing import Dict, Mapping, Optional, List, Any
from pydantic import Field, ConfigDict
from fast_tradier.models.DataClassModelBase import DataClassModelBase
from fast_tradier.models.ModelBase import NewModelBase

class Leg(NewModelBase):
    id: int
    type: str
    symbol: str
    side: str
    quantity: float
    status: str
    duration: str
    price: Optional[float] = None
    avg_fill_price: Optional[float] = None
    exec_quantity: float = 0
    last_fill_price: Optional[float] = None
    last_fill_quantity: float = 0
    remaining_quantity: float = 0
    create_date: str
    transaction_date: str
    # class: str
    class_: Optional[str] = Field(..., alias="class")
    option_symbol: Optional[str] = None
    model_config = ConfigDict(populate_by_name=True)
    
    def to_json(self) -> Mapping[str, Any]:
        result = self.model_dump(by_alias=True)
        return result

class AccountOrder(NewModelBase):
    id: int
    # type: str
    type_: str = Field(..., alias="type")
    symbol: str
    side: str
    quantity: float
    status: str
    duration: str
    price: Optional[float] = None
    avg_fill_price: Optional[float] = None
    exec_quantity: Optional[float] = None
    last_fill_price: Optional[float] = None
    last_fill_quantity: Optional[float] = None
    remaining_quantity: Optional[float] = None
    create_date: Optional[str] = None
    transaction_date: Optional[str] = None
    # class_type: str
    class_: Optional[str] = Field(..., alias="class")
    num_legs: Optional[int] = None
    strategy: Optional[str] = None
    leg: Optional[List[Leg]] = []
    stop_price: Optional[float] = None
    reason_description: Optional[str] = None # rejection details
    tag: Optional[str] = None
    model_config = ConfigDict(populate_by_name=True)

    def to_json(self) -> Mapping[str, Any]:
        result = self.model_dump(by_alias=True)
        return result


from dataclasses import dataclass
from typing import Dict, Optional

from fast_tradier.models.DataClassModelBase import DataClassModelBase
from fast_tradier.models.ModelBase import NewModelBase

class Cash(NewModelBase):
    cash_available: float
    sweep: float
    unsettled_funds: float

class Margin(NewModelBase):
    fed_call: float
    maintenance_call: float
    option_buying_power: float
    stock_buying_power: float
    stock_short_value: float
    sweep: float

class Pdt(NewModelBase):
    fed_call: int
    maintenance_call: float
    option_buying_power: float
    stock_buying_power: float
    stock_short_value: float

class AccountBalance(NewModelBase):
    option_short_value: Optional[float] = None
    total_equity: float
    account_number: str
    account_type: str
    close_pl: float
    current_requirement: float
    equity: int
    long_market_value: float
    market_value: float
    open_pl: float
    option_long_value: float
    option_requirement: int
    pending_orders_count: int
    short_market_value: float
    stock_long_value: float
    total_cash: float
    uncleared_funds: float
    pending_cash: float
    margin: Optional[Margin]  = None
    cash: Optional[Cash] = None
    pdt: Optional[Pdt] = None


from dataclasses import dataclass
from typing import Dict, Optional, Mapping, Any
from pydantic import ConfigDict, Field

from fast_tradier.models.DataClassModelBase import DataClassModelBase
from fast_tradier.models.ModelBase import NewModelBase

class Hlocv(NewModelBase):
    date: str
    open_price: Optional[float] = Field(..., alias="open")
    high: Optional[float] = None
    low: Optional[float] = None
    close: Optional[float] = None
    volume: Optional[int] = None
    model_config = ConfigDict(use_enum_values=True, populate_by_name=True)


from setuptools import setup

with open("README.md") as file:
    long_description = file.read()

setup(
    name="fast-tradier-client",
    version="0.3.1",
    description="Fast Tradier Client in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Tony W",
    author_email="ivytony@gmail.com",
    url="https://pypi.org/project/fast-tradier-client/",
    keywords=["python", "fast-tradier-client", "tradier"],
)


from dataclasses import dataclass
from typing import Dict, Optional
from fast_tradier.models.ModelBase import NewModelBase
from fast_tradier.models.DataClassModelBase import DataClassModelBase

class Position(NewModelBase):
    cost_basis: Optional[float] = None
    date_acquired: Optional[str] = None
    id: int
    quantity: Optional[float] = None
    symbol: str


from typing import Dict, Tuple, Optional, Mapping, Any
from pydantic import Field, ConfigDict
from fast_tradier.models.ModelBase import ModelBase
from fast_tradier.models.ModelBase import NewModelBase

class TradierQuote(NewModelBase):
    symbol: str
    type_: str = Field(..., alias="type")
    open_price: Optional[float] = Field(..., alias="open")
    high: float
    low: float
    close: float
    volume: float
    bid: Optional[float] = None
    ask: Optional[float] = None
    last_price: Optional[float] = None
    model_config = ConfigDict(use_enum_values=True, populate_by_name=True)

    @property
    def ohlcv(self) -> Tuple:
        return self.open_price, self.high, self.low, self.close, self.volume

    @property
    def quote_type(self) -> str:
        return self.type_
    
    @property
    def mid(self) -> Optional[float]:
        return None if self.bid is None or self.ask is None else (self.bid + self.ask) / 2.0